#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Author: Benjamin Vial
# License: MIT

"""
Sources.
"""

import numpy as np
import sympy as sp
from scipy.constants import c, epsilon_0, mu_0
from sympy.vector import CoordSys3D

from .complex import Complex, Constant, as_tensor, dolfin
from .complex import grad as gradc

_COORD = CoordSys3D("N")


def _vector(components):
    return (
        components[0] * _COORD.i + components[1] * _COORD.j + components[2] * _COORD.k
    )


_x = sp.symbols("x[0] x[1] x[2]", real=True)
_X = _vector(np.array(_x))


def _expression2complex_2d(expr, **kwargs):
    re, im = (p.subs(_x[2], 0) for p in expr.as_real_imag())
    dexpr = [dolfin.Expression(sp.printing.ccode(p), **kwargs) for p in (re, im)]
    return Complex(*dexpr)


def plane_wave_2d(lambda0, theta, amplitude=1, degree=1, domain=None):
    k0 = 2 * np.pi / lambda0
    K = k0 * np.array((np.cos(theta), np.sin(theta)))
    K_ = _vector(sp.symbols("kx, ky, 0", real=True))
    expr = amplitude * sp.exp(1j * K_.dot(_X))
    return _expression2complex_2d(expr, kx=K[0], ky=K[1], degree=degree, domain=domain)


def plane_wave_3d(lambda0, theta, phi, psi, amplitude=1, degree=1, domain=None):

    cx = np.cos(psi) * np.cos(theta) * np.cos(phi) - np.sin(psi) * np.sin(phi)
    cy = np.cos(psi) * np.cos(theta) * np.sin(phi) + np.sin(psi) * np.cos(phi)
    cz = -np.cos(psi) * np.sin(theta)

    k0 = 2 * np.pi / lambda0
    K = k0 * np.array(
        (
            np.sin(theta) * np.cos(phi),
            np.sin(theta) * np.sin(phi),
            np.cos(theta),
        )
    )
    K_ = _vector(sp.symbols("kx, ky, kz", real=True))

    Propp = amplitude * sp.exp(1j * K_.dot(_X))

    code = [sp.printing.ccode(p) for p in Propp.as_real_imag()]
    prop = dolfin.Expression(
        code, kx=K[0], ky=K[1], kz=K[2], degree=degree, domain=domain
    )
    C = dolfin.as_tensor([cx, cy, cz])
    return Complex(prop[0] * C, prop[1] * C)


def green_function_2d(lambda0, xs, ys, amplitude=1, degree=1, domain=None):
    Xs = _vector(sp.symbols("xs, ys, 0", real=True))
    k0 = sp.symbols("k0", real=True)
    Xshift = _X - Xs
    rho = sp.sqrt(Xshift.dot(Xshift))
    rho = rho.subs(_x[2], 0)
    kr = k0 * rho
    k0_ = 2 * np.pi / lambda0
    KR = dolfin.Expression(
        sp.printing.ccode(kr), k0=k0_, xs=xs, ys=ys, degree=degree, domain=domain
    )
    return (
        -1
        / 4
        * Complex(dolfin.bessel_Y(0, KR), dolfin.bessel_J(0, KR))
        * Constant(amplitude)
    )


def field_stack_2D(phi, alpha, beta, yshift=0, degree=1, domain=None):
    alpha0_re, alpha0_im, beta0_re, beta0_im = sp.symbols(
        "alpha0_re,alpha0_im,beta0_re,beta0_im", real=True
    )
    alpha0 = alpha0_re + 1j * alpha0_im
    beta0 = beta0_re + 1j * beta0_im
    Kplus = _vector((alpha0, beta0, 0))
    Kminus = _vector((alpha0, -beta0, 0))
    deltaY = _vector((0, sp.symbols("yshift", real=True), 0))
    pw = lambda K: sp.exp(1j * K.dot(_X - deltaY))
    phi_plus_re, phi_plus_im = sp.symbols("phi_plus_re,phi_plus_im", real=True)
    phi_minus_re, phi_minus_im = sp.symbols("phi_minus_re,phi_minus_im", real=True)
    phi_plus = phi_plus_re + 1j * phi_plus_im
    phi_minus = phi_minus_re + 1j * phi_minus_im
    field = phi_plus * pw(Kplus) + phi_minus * pw(Kminus)

    expr = _expression2complex_2d(
        field,
        alpha0_re=alpha.real,
        alpha0_im=alpha.imag,
        beta0_re=beta.real,
        beta0_im=beta.imag,
        phi_plus_re=phi[0].real,
        phi_plus_im=phi[0].imag,
        phi_minus_re=phi[1].real,
        phi_minus_im=phi[1].imag,
        yshift=yshift,
        degree=degree,
        domain=domain,
    )
    return expr


def field_stack_3D(phi, alpha, beta, gamma, zshift=0, degree=1, domain=None):
    alpha0_re, alpha0_im, beta0_re, beta0_im, gamma0_re, gamma0_im = sp.symbols(
        "alpha0_re, alpha0_im, beta0_re, beta0_im, gamma0_re, gamma0_im", real=True
    )
    alpha0 = alpha0_re + 1j * alpha0_im
    beta0 = beta0_re + 1j * beta0_im
    gamma0 = gamma0_re + 1j * gamma0_im
    Kplus = _vector((alpha0, beta0, gamma0))
    Kminus = _vector((alpha0, beta0, -gamma0))
    deltaZ = _vector((0, 0, sp.symbols("zshift", real=True)))
    pw = lambda K: sp.exp(1j * K.dot(_X - deltaZ))
    fields = []
    for comp in ["x", "y", "z"]:
        phi_plus_re, phi_plus_im = sp.symbols(
            f"phi_plus_{comp}_re,phi_plus_{comp}_im", real=True
        )
        phi_minus_re, phi_minus_im = sp.symbols(
            f"phi_minus_{comp}_re,phi_minus_{comp}_im", real=True
        )
        phi_plus = phi_plus_re + 1j * phi_plus_im
        phi_minus = phi_minus_re + 1j * phi_minus_im
        field = phi_plus * pw(Kplus) + phi_minus * pw(Kminus)
        fields.append(field)
    code = [[sp.printing.ccode(p) for p in f.as_real_imag()] for f in fields]
    code = np.ravel(code).tolist()
    expr = [
        dolfin.Expression(
            c,
            alpha0_re=alpha.real,
            alpha0_im=alpha.imag,
            beta0_re=beta.real,
            beta0_im=beta.imag,
            gamma0_re=gamma.real,
            gamma0_im=gamma.imag,
            phi_plus_x_re=phi[0].real,
            phi_plus_x_im=phi[0].imag,
            phi_minus_x_re=phi[1].real,
            phi_minus_x_im=phi[1].imag,
            phi_plus_y_re=phi[2].real,
            phi_plus_y_im=phi[2].imag,
            phi_minus_y_re=phi[3].real,
            phi_minus_y_im=phi[3].imag,
            phi_plus_z_re=phi[4].real,
            phi_plus_z_im=phi[4].imag,
            phi_minus_z_re=phi[5].real,
            phi_minus_z_im=phi[5].imag,
            zshift=zshift,
            degree=degree,
            domain=domain,
        )
        for c in code
    ]

    return Complex(
        as_tensor([expr[0], expr[2], expr[4]]), as_tensor([expr[1], expr[3], expr[5]])
    )


class Source:
    def __init__(self, wavelength, dim, degree=1, domain=None):
        self.wavelength = wavelength
        self.dim = dim
        self.degree = degree
        self.domain = domain

    @property
    def wavenumber(self):
        return 2 * np.pi / self.wavelength

    @property
    def pulsation(self):
        return self.wavenumber * c

    @property
    def frequency(self):
        return c / self.wavelength


class PlaneWave(Source):
    def __init__(self, wavelength, angle, dim, amplitude=1, degree=1, domain=None):
        super().__init__(wavelength, dim, degree=degree, domain=domain)
        self.angle = angle
        self.amplitude = amplitude

    @property
    def expression(self):
        if self.dim == 2:
            _expression = plane_wave_2d(
                self.wavelength,
                self.angle,
                amplitude=self.amplitude,
                degree=self.degree,
                domain=self.domain,
            )
        else:
            _expression = plane_wave_3d(
                self.wavelength,
                *self.angle,
                amplitude=self.amplitude,
                degree=self.degree,
                domain=self.domain,
            )
        return _expression


class LineSource(Source):
    def __init__(self, wavelength, position, amplitude=1, degree=1, domain=None):
        super().__init__(wavelength, dim=2, degree=degree, domain=domain)
        self.position = position
        self.amplitude = amplitude

    @property
    def expression(self):
        _expression = green_function_2d(
            self.wavelength,
            *self.position,
            amplitude=self.amplitude,
            degree=self.degree,
            domain=self.domain,
        )
        return _expression


#
# mesh = dolfin.UnitSquareMesh(50, 50)
# pw = PlaneWave(0.3, 0*np.pi/2, 2, degree=2)
#
# # pw = plane_wave_2d(0.1, 0, degree=2)
#
# from gyptis.plot import *
#
# plt.ion()
#
# W = dolfin.FunctionSpace(mesh,"CG",2)
#
# Wc = ComplexFunctionSpace(mesh,"CG",2)
#
# # pwexp = project(pw.expression.real, W)
# # pwexp = project(pw.expression, W)
#
# plt.close("all")
# plt.figure()
# cb = dolfin.plot(pw.expression.real, mesh=mesh)
# cb = dolfin.plot(pw.gradient[0].real, mesh=mesh)
# plt.colorbar(cb)
#
#
# plt.figure()
# pw.wavelength=0.1
# pw.angle=0.2
# cb = dolfin.plot(pw.expression.real, mesh=mesh)
# cb = dolfin.plot(pw.gradient[0].real, mesh=mesh)
# plt.colorbar(cb)
#
# #
# # cdsc
# # plt.clf()
# #
# # cb = dolfin.plot(pw.gradient[0].real, mesh=mesh)
# # plt.colorbar(cb)
# # # dolfin.plot(pw.gradient[0].imag, mesh=mesh)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Author: Benjamin Vial
# License: MIT


"""
Geometry definition using Gmsh api.
For more information see Gmsh's `documentation <https://gmsh.info/doc/texinfo/gmsh.html>`_
"""

import numbers
import re
import sys
import tempfile
from functools import wraps

import gmsh
import numpy as np

from . import dolfin
from .measure import Measure
from .mesh import read_mesh

_geometry_module = sys.modules[__name__]

geo = gmsh.model.geo
occ = gmsh.model.occ
setnum = gmsh.option.setNumber

gmsh_options = gmsh.option

# def copy_docstring_from(source):
#     def wrapper(func):
#         func.__doc__ = source.__doc__
#         return func
#     return wrapper


def _set_opt_gmsh(name, value):
    if isinstance(value, str):
        return gmsh_options.setString(name, value)
    elif isinstance(value, numbers.Number) or isinstance(value, bool):
        if isinstance(value, bool):
            value = int(value)
        return gmsh_options.setNumber(name, value)
    else:
        raise ValueError("value must be string or number")


def _get_opt_gmsh(name):
    try:
        return gmsh_options.getNumber(name)
    except:
        return gmsh_options.getString(name)


setattr(gmsh_options, "set", _set_opt_gmsh)
setattr(gmsh_options, "get", _get_opt_gmsh)


def _add_method(cls, func, name):
    @wraps(func)
    def wrapper(*args, sync=True, **kwargs):
        out = func(*args, **kwargs)
        if sync:
            occ.synchronize()
        return out

    setattr(cls, name, wrapper)
    return func


def _dimtag(tag, dim=3):
    if not isinstance(tag, list):
        tag = list([tag])
    return [(dim, t) for t in tag]


def _get_bnd(id, dim):
    out = gmsh.model.getBoundary(_dimtag(id, dim=dim), False, False, False)
    return [b[1] for b in out]


def _convert_name(name):
    return re.sub(r"(?<!^)(?=[A-Z])", "_", name).lower()


def _make_func(cls, name):
    class_name = getattr(_geometry_module, "_" + name)

    @wraps(class_name.__init__)
    def wrapper(self, *args, **kwargs):
        return class_name(cls, self, *args, **kwargs)

    setattr(cls, name, wrapper)
    # getattr(cls, name).__doc__ = class_name.__init__.__doc__
    return wrapper


class GeometryPrimitive(object):
    def __init__(
        self, geometry, add_function=None, id=-1, dim=2, name=None, physical_name=None
    ):
        self.geometry = geometry
        self.add_function = add_function
        self.id = id
        self.name = name
        self.dim = dim
        self._physical_name = physical_name
        # self.physical_name = physical_name

    def _get_geometry_method(self, m):
        return getattr(self.geometry, m)

    def _basic_operation(self, method, other):
        op = self._get_geometry_method(method)
        id = op(self.id, other.id)
        new = GeometryPrimitive(self.geometry)
        new.id = id
        return new

    def get_boundaries(self, physical=False):
        bnds = self._get_geometry_method("get_boundaries")(
            self.id, self.dim, physical=physical
        )
        new = [GeometryPrimitive(self.geometry, dim=self.dim - 1, id=id) for id in bnds]
        return new

    def add(self, *args, **kwargs):
        if self.add_function is not None:
            return self._get_geometry_method(self.add_function)(*args, **kwargs)

    def __len__(self):
        return len(self.id)

    def __add__(self, other):
        return self._basic_operation("fuse", other)

    def __sub__(self, other):
        return self._basic_operation("cut", other)

    @property
    def physical_name(self):
        return self._physical_name

    @physical_name.setter
    def physical_name(self, name):
        self.geometry.add_physical(self.id, name)
        self._physical_name = name

    def add_physical(self, name):
        self.physical_name = name

    def __lshift__(self, name):
        self.add_physical(name)

    def __truediv__(self, other):
        op = self._get_geometry_method("fragment")
        ids, map = op(self.id, other.id, map=True)
        new = []
        for id in ids:
            p = GeometryPrimitive(self.geometry)
            p.id = id
            new.append(p)
        return new

    def rotate(self, point, axis, angle):
        new = GeometryPrimitive(self.geometry)
        self._get_geometry_method("rotate")(self.id, point, axis, angle)
        new.id = self.id
        return new

    def extrude(self, vector, num_el=[], heights=[], recombine=False):
        new = GeometryPrimitive(self.geometry)
        dimtags = self.geometry.dimtag(self.id)

        self._get_geometry_method("extrude")(
            dimtags, *vector, num_el, heights, recombine
        )
        new.id = self.id
        new.dim = self.dim + 1
        return new

    def set_size(self, size, **kwargs):
        self.geometry.set_size(self.id, size, **kwargs)

    def __or__(self, size):
        return self.set_size(size)

    def __repr__(self):
        return f"GeometryPrimitive(id={self.id}, dim={self.dim})"

    # def __rshift__(self, other):
    #
    #     return _basic_operation("fragment", self, other)


class _Circle(GeometryPrimitive):
    def __init__(self, geometry, center, radius, surface=True, name=None, **kwargs):
        """
        Circle(center, radius, surface=True, name=None, **kwargs)

        Create and add a circle.

        Parameters
        ----------
        center : tuple
            Center (x,y,z).
        radius : float
            Radius.
        surface : bool
            Wether to create a plane surface (the default is True).
        name : str
            Name of the object (the default is None).

        Returns
        -------
        Circle
            A circle object.

        """
        super().__init__(geometry, "add_circle")
        self.radius = radius
        self.center = center
        self.surface = surface
        self.name = name
        self.add(**kwargs)

    def __repr__(self):
        return f"Circle(r={self.radius}, c={self.center})"

    def add(self, **kwargs):
        self.id = super().add(*self.center, self.radius, surface=self.surface, **kwargs)


class _Rectangle(GeometryPrimitive):
    def __init__(self, geometry, corner, size, corner_radius=0.0, name=None, **kwargs):
        """
        Rectangle(corner, size, corner_radius=0.0, name=None)

        Create and add a rectangle.

        Parameters
        ----------
        corner : tuple
            Corner (x,y,z).
        size : tuple
            Size (Lx, Ly).
        corner_radius : float
            Radius of corners (the default is 0.0).
        name : str
            Name of the object (the default is None).

        Returns
        -------
        Rectangle
            A rectangle object.

        """
        super().__init__(geometry, "add_rectangle")
        self.size = size
        self.corner = corner
        self.corner_radius = corner_radius
        self.name = name
        self.add(**kwargs)

    def add(self, **kwargs):
        self.id = super().add(
            *self.corner, *self.size, roundedRadius=self.corner_radius, **kwargs
        )


class _Ellipse(GeometryPrimitive):
    def __init__(self, geometry, center, radii, surface=True, name=None, **kwargs):
        """
        Ellipse(center, radii, surface=True, name=None, **kwargs)

        Create and add an ellipse.

        Parameters
        ----------
        center : tuple
            Center (x,y,z).
        radii : tuple
            Axes of the ellipse (Lx,Ly).
        surface : bool
            Wether to create a plane surface (the default is True).
        name : str
            Name of the object (the default is None).

        Returns
        -------
        Circle
            A circle object.

        """
        super().__init__(geometry, "add_ellipse")
        self.radii = radii
        self.center = center
        self.surface = surface
        self.name = name
        self.add(**kwargs)

    def __repr__(self):
        return f"Circle(r={self.radius}, c={self.center})"

    def add(self, **kwargs):
        self.id = super().add(*self.center, *self.radii, surface=self.surface, **kwargs)


class Geometry(object):
    """Base class for geometry models.

    Parameters
    ----------
    model_name : str
        Name of the model
    mesh_name : str
        Name of the mesh file (.msh).
    data_dir : str
        Directory to store generated mesh and data.
    dim : int
        Dimension of the problem.

    Attributes
    ----------
    subdomains : dict
        Dictionary containing mapping from physical domains names to their
        index.
    model_name
    mesh_name
    dim
    data_dir

    """

    def __init__(
        self,
        model_name="geometry",
        mesh_name="mesh.msh",
        data_dir=None,
        dim=3,
        gmsh_args=None,
        kill=True,
        verbose=0,
    ):
        self.model_name = model_name
        self.mesh_name = mesh_name
        self.dim = dim
        self.subdomains = dict(volumes={}, surfaces={}, curves={}, points={})
        self.data_dir = data_dir if data_dir else tempfile.mkdtemp()
        self.occ = occ
        self.mesh_object = {}
        self.measure = {}
        self.mesh = {}
        self.markers = {}
        self.verbose = verbose

        for object_name in dir(occ):
            if (
                not object_name.startswith("__")
                and object_name != "mesh"
                and object_name not in dir(self)
            ):
                bound_method = getattr(occ, object_name)
                name = _convert_name(bound_method.__name__)
                _add_method(self, bound_method, name)

        self._gmsh_add_ellipse = self.add_ellipse
        del self.add_ellipse
        self._gmsh_add_circle = self.add_circle
        del self.add_circle
        self._gmsh_add_spline = self.add_spline
        del self.add_spline

        if kill:
            try:
                gmsh.finalize()
            except:
                pass

        self.gmsh_args = gmsh_args
        if gmsh_args is not None:
            gmsh.initialize(self.gmsh_args)
        else:
            gmsh.initialize()

        gmsh_options.set("General.Verbosity", self.verbose)

    @wraps(_Circle.__init__)
    def Circle(self, *args, **kwargs):
        return _Circle(self, *args, **kwargs)

    @wraps(_Rectangle.__init__)
    def Rectangle(self, *args, **kwargs):
        return _Rectangle(self, *args, **kwargs)

    @wraps(_Ellipse.__init__)
    def Ellipse(self, *args, **kwargs):
        return _Ellipse(self, *args, **kwargs)

    # def add_rectangle(self, *args, **kwargs):
    #     return self.add_rectangle(*args, **kwargs)

    def rotate(self, tag, point, axis, angle, dim=None):
        dt = self.dimtag(tag, dim=dim)
        return occ.rotate(dt, *point, *axis, angle)

    def add_physical(self, id, name, dim=None):
        """Add a physical domain.

        Parameters
        ----------
        id : int or list of int
            The identifiant(s) of elementary entities making the physical domain.
        name : str
            Name of the domain.
        dim : int
            Dimension.
        """
        dim = dim if dim else self.dim
        dicname = list(self.subdomains)[3 - dim]
        if not isinstance(id, list):
            id = list([id])
        num = gmsh.model.addPhysicalGroup(dim, id)
        self.subdomains[dicname][name] = num
        gmsh.model.removePhysicalName(name)
        gmsh.model.setPhysicalName(dim, self.subdomains[dicname][name], name)
        return num

    def dimtag(self, id, dim=None):
        """Convert an integer or list of integer to gmsh DimTag notation.

        Parameters
        ----------
        id : int or list of int
            Label or list of labels.
        dim : type
            Dimension.

        Returns
        -------
        int or list of int
            A tuple (dim, tag) or list of such tuples (gmsh DimTag notation).

        """
        dim = dim or self.dim
        return _dimtag(id, dim=dim)

    # def add_circle(self,x, y, z, ax, ay,**kwargs):
    #     ell = self._gmsh_add_ellipse(x, y, z, ax, ay,**kwargs)
    #     ell = self.add_curve_loop([ell])
    #     return self.add_plane_surface([ell])

    def add_circle(self, x, y, z, r, surface=True, **kwargs):
        if surface:
            circ = self._gmsh_add_circle(x, y, z, r, **kwargs)
            circ = self.add_curve_loop([circ])
            return self.add_plane_surface([circ])
        else:
            return self._gmsh_add_circle(x, y, z, r, **kwargs)

    def add_ellipse(self, x, y, z, ax, ay, surface=True, **kwargs):
        if ax == ay:
            return self.add_circle(x, y, z, ax, surface=surface, **kwargs)
        elif ax < ay:
            ell = self.add_ellipse(x, y, z, ay, ax, surface=surface, **kwargs)
            self.rotate(ell, (x, y, z), (0, 0, 1), np.pi / 2, dim=2)
            return ell
        else:
            if surface:
                ell = self._gmsh_add_ellipse(x, y, z, ax, ay, **kwargs)
                ell = self.add_curve_loop([ell])
                return self.add_plane_surface([ell])
            else:
                return self._gmsh_add_ellipse(x, y, z, ax, ay, **kwargs)

    def add_spline(self, points, mesh_size=0.0, surface=True, **kwargs):
        """Adds a spline.

        Parameters
        ----------
        points : array of shape (Npoints,3)
            Corrdinates of the points.
        mesh_size : float
            Mesh sizes at points (the default is 0.0).
        surface : type
            If True, creates a plane surface (the default is True).

        Returns
        -------
        int
            The tag of the spline.

        """
        dt = []
        for p in points:
            dt.append(self.add_point(*p, meshSize=mesh_size))

        if np.allclose(points[0], points[-1]):
            dt[-1] = dt[0]

        if surface:
            spl = self._gmsh_add_spline(dt, **kwargs)
            spl = self.add_curve_loop([spl])
            return self.add_plane_surface([spl])
        else:
            return self._gmsh_add_spline(dt, **kwargs)

    def fragment(self, id1, id2, dim1=None, dim2=None, sync=True, map=False, **kwargs):
        dim1 = dim1 if dim1 else self.dim
        dim2 = dim2 if dim2 else self.dim
        a1 = self.dimtag(id1, dim1)
        a2 = self.dimtag(id2, dim2)
        dimtags, mapping = occ.fragment(a1, a2, **kwargs)
        if sync:
            occ.synchronize()
        tags = [_[1] for _ in dimtags]
        if map:
            return tags, mapping
        else:
            return tags

    def cut(self, id1, id2, dim1=None, dim2=None, sync=True, **kwargs):
        dim1 = dim1 if dim1 else self.dim
        dim2 = dim2 if dim2 else self.dim
        a1 = self.dimtag(id1, dim1)
        a2 = self.dimtag(id2, dim2)
        ov, ovv = occ.cut(a1, a2, **kwargs)
        if sync:
            occ.synchronize()
        return [o[1] for o in ov]

    def fuse(self, id1, id2, dim1=None, dim2=None, sync=True):
        dim1 = dim1 if dim1 else self.dim
        dim2 = dim2 if dim2 else self.dim
        a1 = self.dimtag(id1, dim1)
        a2 = self.dimtag(id2, dim2)
        ov, ovv = occ.fuse(a1, a2)
        if sync:
            occ.synchronize()
        return [o[1] for o in ov]

    def get_boundaries(self, id, dim=None, physical=True):
        dim = dim if dim else self.dim
        if isinstance(id, str):
            if dim == 3:
                type_entity = "volumes"
            elif dim == 2:
                type_entity = "surfaces"
            else:
                type_entity = "curves"
            id = self.subdomains[type_entity][id]

            n = gmsh.model.getEntitiesForPhysicalGroup(dim, id)
            bnds = [_get_bnd(n_, dim=dim) for n_ in n]
            bnds = [item for sublist in bnds for item in sublist]
            return list(dict.fromkeys(bnds))
        else:
            if physical:
                n = gmsh.model.getEntitiesForPhysicalGroup(dim, id)[0]
            else:
                n = id
            return _get_bnd(n, dim=dim)

    def _set_size(self, id, s, dim=None):
        dim = dim if dim else self.dim
        p = gmsh.model.getBoundary(
            self.dimtag(id, dim=dim), False, False, True
        )  # Get all points
        gmsh.model.mesh.setSize(p, s)

    def _check_subdomains(self):
        groups = gmsh.model.getPhysicalGroups()
        names = [gmsh.model.getPhysicalName(*g) for g in groups]
        for subtype, subitems in self.subdomains.items():
            for id in subitems.copy().keys():
                if id not in names:
                    subitems.pop(id)

    def set_mesh_size(self, params, dim=None):
        dim = dim if dim else self.dim
        if dim == 3:
            type_entity = "volumes"
        elif dim == 2:
            type_entity = "surfaces"
        elif dim == 1:
            type_entity = "curves"
        elif dim == 0:
            type_entity = "points"

        # revert sort so that smaller sizes are set last
        params = dict(
            sorted(params.items(), key=lambda item: float(item[1]), reverse=True)
        )

        for id, p in params.items():
            if isinstance(id, str):
                num = self.subdomains[type_entity][id]
                n = gmsh.model.getEntitiesForPhysicalGroup(dim, num)
                for n_ in n:
                    self._set_size(n_, p, dim=dim)
            else:
                self._set_size(id, p, dim=dim)

    def set_size(self, id, s, dim=None):
        if hasattr(id, "__len__") and not isinstance(id, str):
            for i, id_ in enumerate(id):
                if hasattr(s, "__len__"):
                    s_ = s[i]
                else:
                    s_ = s
                params = {id_: s_}
                self.set_mesh_size(params, dim=dim)
        else:
            self.set_mesh_size({id: s}, dim=dim)

    def read_mesh_info(self):
        if self.dim == 2:
            marker_dim = "triangle"
            sub_dim = "surfaces"
            marker_dim_minus_1 = "line"
            sub_dim_dim_minus_1 = "curves"
        else:
            marker_dim = "tetra"
            sub_dim = "volumes"
            marker_dim_minus_1 = "triangle"
            sub_dim_dim_minus_1 = "surfaces"

        self.measure["dx"] = Measure(
            "dx",
            domain=self.mesh_object["mesh"],
            subdomain_data=self.mesh_object["markers"][marker_dim],
            subdomain_dict=self.subdomains[sub_dim],
        )

        # exterior_facets
        if (marker_dim_minus_1 in self.mesh_object["markers"].keys()) and (
            sub_dim_dim_minus_1 in self.subdomains.keys()
        ):
            self.measure["ds"] = Measure(
                "ds",
                domain=self.mesh_object["mesh"],
                subdomain_data=self.mesh_object["markers"][marker_dim_minus_1],
                subdomain_dict=self.subdomains[sub_dim_dim_minus_1],
            )

            # interior_facets

            self.measure["dS"] = Measure(
                "dS",
                domain=self.mesh_object["mesh"],
                subdomain_data=self.mesh_object["markers"][marker_dim_minus_1],
                subdomain_dict=self.subdomains[sub_dim_dim_minus_1],
            )
        else:
            self.measure["ds"] = None
            self.measure["dS"] = None

        self.mesh = self.mesh_object["mesh"]
        self.markers = self.mesh_object["markers"]

        if self.dim == 2:
            self.domains = self.subdomains["surfaces"]
            self.boundaries = self.subdomains["curves"]
            self.lines = {}
            self.points = self.subdomains["points"]
            self.markers = self.mesh_object["markers"]["triangle"]
            self.boundary_markers = (
                self.mesh_object["markers"]["line"] if self.boundaries else []
            )

        else:
            self.domains = self.subdomains["volumes"]
            self.boundaries = self.subdomains["surfaces"]
            self.lines = self.subdomains["curves"]
            self.points = self.subdomains["points"]
            self.markers = self.mesh_object["markers"]["tetra"]
            self.boundary_markers = (
                self.mesh_object["markers"]["triangle"] if self.boundaries else []
            )

        self.unit_normal_vector = dolfin.FacetNormal(self.mesh)

    @property
    def msh_file(self):
        return f"{self.data_dir}/{self.mesh_name}"

    def build(
        self,
        interactive=False,
        generate_mesh=True,
        write_mesh=True,
        read_info=True,
        read_mesh=True,
        finalize=True,
        check_subdomains=True,
    ):
        if check_subdomains:
            self._check_subdomains()

        self.mesh_object = self.generate_mesh(
            generate=generate_mesh, write=write_mesh, read=read_mesh
        )

        if read_info:
            self.read_mesh_info()

        if interactive:
            gmsh.fltk.run()
        if finalize:
            gmsh.finalize()
        return self.mesh_object

    def read_mesh_file(self):
        return read_mesh(self.msh_file, data_dir=self.data_dir, dim=self.dim)

    def generate_mesh(self, generate=True, write=True, read=True):
        if generate:
            gmsh.model.mesh.generate(self.dim)
        if write:
            gmsh.write(self.msh_file)
        if read:
            return self.read_mesh_file()

    def plot_mesh(self, **kwargs):
        return dolfin.plot(self.mesh, **kwargs)

    def plot_subdomains(self, **kwargs):
        from .plot import plot_subdomains

        return plot_subdomains(self.markers, **kwargs)

    def add(self, primittive):
        primittive.add(self)
        ids = []
        for p in primittive.parents:
            self.add(p)
            ids.append(p.id)
        for op in primittive.operations:
            IDs = op(self, *ids)
            primittive.id = IDs



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Author: Benjamin Vial
# License: MIT

"""
Finite element weak formulations.
"""

from abc import ABC, abstractmethod

import numpy as np

from . import dolfin
from .bc import *
from .complex import *
from .source import PlaneWave
from .stack import make_stack


class Formulation(ABC):
    def __init__(
        self,
        geometry,
        coefficients,
        function_space,
        source=None,
        boundary_conditions={},
    ):
        self.geometry = geometry
        self.coefficients = coefficients
        self.function_space = function_space
        self.source = source
        self.trial = TrialFunction(self.function_space)
        self.test = TestFunction(self.function_space)
        self.boundary_conditions = boundary_conditions

        self.measure = geometry.measure
        self.dx = self.measure["dx"]
        self.ds = self.measure["ds"]
        self.dS = self.measure["dS"]

        self.element = self.function_space.split()[0].ufl_element()
        self.real_function_space = dolfin.FunctionSpace(
            self.geometry.mesh, self.element.family(), self.element.degree()
        )

    def build_lhs(self):
        self.lhs = dolfin.lhs(self.weak)
        return self.lhs

    def build_rhs(self):
        self.rhs = dolfin.rhs(self.weak)
        if self.rhs.empty():
            if self.element.value_size() == 3:
                dummy_vect = as_vector(
                    [dolfin.DOLFIN_EPS, dolfin.DOLFIN_EPS, dolfin.DOLFIN_EPS]
                )
                dummy_form = dot(dummy_vect, self.trial) * self.dx
            else:
                dummy_form = dolfin.DOLFIN_EPS * self.trial * self.dx
            self.rhs = dummy_form.real + dummy_form.imag
        return self.rhs

    @abstractmethod
    def weak(self):
        pass

    # @abstractmethod
    # def build_lhs(self):
    #     pass
    #
    # @abstractmethod
    # def build_rhs(self):
    #     pass

    @abstractmethod
    def build_boundary_conditions(self):
        pass


def _find_domains_function(coeffs, list_domains=None):
    dom_function = []
    for coeff in coeffs:
        list_domains = list_domains or list(coeff.dict.keys())
        dom_function += [k for k, v in coeff.dict.items() if hasattr(v, "ufl_shape")]
    dom_function = np.unique(dom_function).tolist()
    dom_no_function = [k for k in list_domains if k not in dom_function]
    return dom_function, dom_no_function


class Maxwell2D(Formulation):
    def __init__(
        self,
        geometry,
        coefficients,
        function_space,
        source=None,
        boundary_conditions={},
        polarization="TE",
        source_domains=[],
        reference=None,
    ):
        super().__init__(
            geometry,
            coefficients,
            function_space,
            source=source,
            boundary_conditions=boundary_conditions,
        )

        self.source_domains = source_domains
        self.reference = reference

        self.epsilon, self.mu = self.coefficients
        self.polarization = polarization

        if self.polarization == "TE":
            self.xi = self.mu.to_xi()
            self.chi = self.epsilon.to_chi()
        else:
            self.xi = self.epsilon.to_xi()
            self.chi = self.mu.to_chi()

        self.pec_boundaries = prepare_boundary_conditions(boundary_conditions)

    def maxwell(self, u, v, xi, chi, domain="everywhere"):
        k0 = Constant(self.source.wavenumber)
        form = -inner(xi * grad(u), grad(v)) + k0 ** 2 * chi * u * v
        return form * self.dx(domain)

    def _weak(self, u, v, u1):
        xi = self.xi.as_subdomain()
        chi = self.chi.as_subdomain()
        xi_a = self.xi.build_annex(
            domains=self.source_domains, reference=self.reference
        ).as_subdomain()
        chi_a = self.chi.build_annex(
            domains=self.source_domains, reference=self.reference
        ).as_subdomain()

        xi_dict = self.xi.as_property()
        chi_dict = self.chi.as_property()
        xi_a_dict = self.xi.build_annex(
            domains=self.source_domains, reference=self.reference
        ).as_property()
        chi_a_dict = self.chi.build_annex(
            domains=self.source_domains, reference=self.reference
        ).as_property()

        dom_func, dom_no_func = _find_domains_function((self.xi, self.chi))
        source_dom_func, source_dom_no_func = _find_domains_function(
            (self.xi, self.chi), self.source_domains
        )

        form = self.maxwell(u, v, xi, chi, domain=dom_no_func)
        for dom in dom_func:
            form += self.maxwell(u, v, xi_dict[dom], chi_dict[dom], domain=dom)

        if self.source_domains != []:
            if source_dom_no_func != []:
                form += self.maxwell(
                    u1, v, xi - xi_a, chi - chi_a, domain=source_dom_no_func
                )
            for dom in source_dom_func:
                form += self.maxwell(
                    u1,
                    v,
                    xi_dict[dom] - xi_a_dict[dom],
                    chi_dict[dom] - chi_a_dict[dom],
                    domain=dom,
                )
        if self.polarization == "TM":
            for bnd in self.pec_boundaries:
                normal = self.geometry.unit_normal_vector
                form -= dot(xi * (grad(u1) * v), normal) * self.ds(bnd)
        weak = form.real + form.imag
        return weak

    @property
    def weak(self):
        u1 = self.source.expression
        u = self.trial
        v = self.test
        return self._weak(u, v, u1)

    def build_pec_boundary_conditions(self, applied_function):
        if self.polarization == "TE" and self.pec_boundaries != []:
            ## FIXME: project is slow, avoid it.
            applied_function = project(applied_function, self.real_function_space)
            _boundary_conditions = build_pec_boundary_conditions(
                self.pec_boundaries,
                self.geometry,
                self.function_space,
                applied_function,
            )
        else:
            _boundary_conditions = []

        return _boundary_conditions

    def build_boundary_conditions(self):
        applied_function = -self.source.expression
        self._boundary_conditions = self.build_pec_boundary_conditions(applied_function)
        return self._boundary_conditions


class Maxwell2DPeriodic(Maxwell2D):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.propagation_vector = self.source.wavenumber * np.array(
            [np.cos(self.source.angle), np.sin(self.source.angle)]
        )
        self.phasor = phasor(
            self.propagation_vector[0],
            direction=0,
            degree=self.source.degree,
            domain=self.geometry.mesh,
        )
        self.annex_field = make_stack(
            self.geometry,
            self.coefficients,
            self.source,
            polarization=self.polarization,
            source_domains=self.source_domains,
            degree=self.source.degree,
            dim=2,
        )

    @property
    def weak(self):
        u1 = self.annex_field["as_subdomain"]["stack"]
        u = self.trial * self.phasor
        v = self.test * self.phasor.conj
        return super()._weak(u, v, u1)

    def build_boundary_conditions(self):
        applied_function = -self.annex_field["as_subdomain"]["stack"] * self.phasor.conj
        self._boundary_conditions = self.build_pec_boundary_conditions(applied_function)
        return self._boundary_conditions


class Maxwell3D(Formulation):
    def __init__(
        self,
        geometry,
        coefficients,
        function_space,
        source=None,
        boundary_conditions={},
        source_domains=[],
        reference=None,
    ):
        super().__init__(
            geometry,
            coefficients,
            function_space,
            source=source,
            boundary_conditions=boundary_conditions,
        )

        self.source_domains = source_domains
        self.reference = reference
        self.epsilon, self.mu = self.coefficients
        self.pec_boundaries = prepare_boundary_conditions(boundary_conditions)

    def maxwell(self, u, v, epsilon, inv_mu, domain="everywhere"):
        k0 = Constant(self.source.wavenumber)
        form = -inner(inv_mu * curl(u), curl(v)) + k0 ** 2 * inner(epsilon * u, v)
        return form * self.dx(domain)

    def _weak(self, u, v, u1):
        epsilon = self.epsilon.as_subdomain()
        inv_mu = self.mu.invert().as_subdomain()

        epsilon_a = self.epsilon.build_annex(
            domains=self.source_domains, reference=self.reference
        ).as_subdomain()
        inv_mu_a = (
            self.mu.invert()
            .build_annex(domains=self.source_domains, reference=self.reference)
            .as_subdomain()
        )
        form = self.maxwell(u, v, epsilon, inv_mu)
        if self.source_domains != []:
            form += self.maxwell(
                u1,
                v,
                epsilon - epsilon_a,
                inv_mu - inv_mu_a,
                domain=self.source_domains,
            )

        # for bnd in self.pec_boundaries:
        #     normal = self.geometry.unit_normal_vector
        #     form -= dot(xi * (grad(u1) * v), normal) * self.ds(bnd)
        weak = form.real + form.imag
        return weak

    @property
    def weak(self):
        u1 = self.source.expression
        u = self.trial
        v = self.test
        return self._weak(u, v, u1)

    def build_pec_boundary_conditions(self, applied_function):
        if self.pec_boundaries != []:
            applied_function = project(applied_function, self.real_function_space)
            _boundary_conditions = build_pec_boundary_conditions(
                self.pec_boundaries,
                self.geometry,
                self.function_space,
                applied_function,
            )
        else:
            _boundary_conditions = []
        return _boundary_conditions

    def build_boundary_conditions(self):
        applied_function = -self.source.expression
        self._boundary_conditions = self.build_pec_boundary_conditions(applied_function)
        return self._boundary_conditions


class Maxwell3DPeriodic(Maxwell3D):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        k0 = self.source.wavenumber
        theta0, phi0 = self.source.angle[0:2]
        alpha0 = k0 * np.cos(theta0) * np.cos(phi0)
        beta0 = k0 * np.cos(theta0) * np.sin(phi0)
        gamma0 = k0 * np.sin(theta0)
        self.propagation_vector = np.array([alpha0, beta0, gamma0])

        self.phasor_vect = [
            phasor(
                self.propagation_vector[i],
                direction=i,
                degree=self.source.degree,
                domain=self.geometry.mesh,
            )
            for i in range(3)
        ]
        self.phasor = self.phasor_vect[0] * self.phasor_vect[1]
        self.annex_field = make_stack(
            self.geometry,
            self.coefficients,
            self.source,
            source_domains=self.source_domains,
            degree=self.source.degree,
            dim=3,
        )

    @property
    def weak(self):
        u1 = self.annex_field["as_subdomain"]["stack"]
        u = self.trial * self.phasor
        v = self.test * self.phasor.conj
        return self._weak(u, v, u1)

    def build_boundary_conditions(self):
        applied_function = -self.annex_field["as_subdomain"]["stack"] * self.phasor.conj
        self._boundary_conditions = self.build_pec_boundary_conditions(applied_function)
        return self._boundary_conditions



"""
Information about the package.
"""

from datetime import date

__version__ = "0.1.2"
__author__ = u"Benjamin Vial"
__author_email__ = "benjamin.vial84@gmail.com"
__copyright__ = u"Copyright (c) 2020-{}, {} <{}>".format(
    date.today().year, __author__, __author_email__
)
__website__ = "https://gyptis.gitlab.io"
__license__ = "License :: OSI Approved :: MIT License"
__status__ = "Development Status :: 1 - Planning"
__audience__ = "Intended Audience :: Science/Research"
__operating_system__ = "Operating System :: OS Independent"
__programming_language__ = [
    "Programming Language :: Python",
    "Programming Language :: Python :: 3",
]
__topic__ = [
    "Topic :: Scientific/Engineering",
    "Topic :: Scientific/Engineering :: Physics",
]
__description__ = "Computational Photonics in Python"



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Author: Benjamin Vial
# License: MIT


from . import dolfin
from .complex import *
from .formulation import Maxwell3D
from .geometry import *
from .materials import *
from .simulation import Simulation
from .source import *


class BoxPML3D(Geometry):
    def __init__(
        self,
        box_size=(1, 1, 1),
        box_center=(0, 0, 0),
        pml_width=(0.2, 0.2, 0.2),
        model_name="3D box with PMLs",
        mesh_name="mesh.msh",
        data_dir=None,
        **kwargs,
    ):
        super().__init__(
            model_name=model_name, mesh_name=mesh_name, data_dir=data_dir, dim=3
        )
        self.box_size = box_size
        self.box_center = box_center
        self.pml_width = pml_width

        def _addbox_center(rect_size):
            corner = -np.array(rect_size) / 2
            corner = tuple(corner)
            return self.add_box(*corner, *rect_size)

        def _translate(tag, t):
            translation = tuple(t)
            self.translate(self.dimtag(tag), *translation)

        def _add_pml(s, t):
            pml = _addbox_center(s)
            _translate(pml, t)
            return pml

        box = _addbox_center(self.box_size)
        T = np.array(self.pml_width) / 2 + np.array(self.box_size) / 2

        s = (self.pml_width[0], self.box_size[1], self.box_size[2])
        t = np.array([T[0], 0, 0])
        pmlxp = _add_pml(s, t)
        pmlxm = _add_pml(s, -t)
        s = (self.box_size[0], self.pml_width[1], self.box_size[2])
        t = np.array([0, T[1], 0])
        pmlyp = _add_pml(s, t)
        pmlym = _add_pml(s, -t)
        s = (self.box_size[0], self.box_size[1], self.pml_width[2])
        t = np.array([0, 0, T[2]])
        pmlzp = _add_pml(s, t)
        pmlzm = _add_pml(s, -t)

        s = (self.pml_width[0], self.pml_width[1], self.box_size[2])

        pmlxypp = _add_pml(s, [T[0], T[1], 0])
        pmlxypm = _add_pml(s, [T[0], -T[1], 0])
        pmlxymp = _add_pml(s, [-T[0], T[1], 0])
        pmlxymm = _add_pml(s, [-T[0], -T[1], 0])

        s = (self.box_size[0], self.pml_width[1], self.pml_width[2])
        pmlyzpp = _add_pml(s, [0, T[1], T[0]])
        pmlyzpm = _add_pml(s, [0, T[1], -T[0]])
        pmlyzmp = _add_pml(s, [0, -T[1], T[0]])
        pmlyzmm = _add_pml(s, [0, -T[1], -T[0]])

        s = (self.pml_width[0], self.box_size[1], self.pml_width[2])
        pmlxzpp = _add_pml(s, [T[0], 0, T[2]])
        pmlxzpm = _add_pml(s, [T[0], 0, -T[2]])
        pmlxzmp = _add_pml(s, [-T[0], 0, T[2]])
        pmlxzmm = _add_pml(s, [-T[0], 0, -T[2]])

        s = (self.pml_width[0], self.pml_width[1], self.pml_width[2])
        pmlxyzppp = _add_pml(s, [T[0], T[1], T[2]])
        pmlxyzppm = _add_pml(s, [T[0], T[1], -T[2]])
        pmlxyzpmp = _add_pml(s, [T[0], -T[1], T[2]])
        pmlxyzpmm = _add_pml(s, [T[0], -T[1], -T[2]])
        pmlxyzmpp = _add_pml(s, [-T[0], T[1], T[2]])
        pmlxyzmpm = _add_pml(s, [-T[0], T[1], -T[2]])
        pmlxyzmmp = _add_pml(s, [-T[0], -T[1], T[2]])
        pmlxyzmmm = _add_pml(s, [-T[0], -T[1], -T[2]])

        pmlx = [pmlxp, pmlxm]
        pmly = [pmlyp, pmlym]
        pmlz = [pmlzp, pmlzm]
        pml1 = pmlx + pmly + pmlz

        pmlxy = [pmlxypp, pmlxypm, pmlxymp, pmlxymm]
        pmlyz = [pmlyzpp, pmlyzpm, pmlyzmp, pmlyzmm]
        pmlxz = [pmlxzpp, pmlxzpm, pmlxzmp, pmlxzmm]
        pml2 = pmlxy + pmlyz + pmlxz

        pml3 = [
            pmlxyzppp,
            pmlxyzppm,
            pmlxyzpmp,
            pmlxyzpmm,
            pmlxyzmpp,
            pmlxyzmpm,
            pmlxyzmmp,
            pmlxyzmmm,
        ]

        self.box = box
        self.pmls = pml1 + pml2 + pml3

        _translate([self.box] + self.pmls, self.box_center)

        self.fragment(self.box, self.pmls)
        #

        self.add_physical(box, "box")
        self.add_physical(pmlx, "pmlx")
        self.add_physical(pmly, "pmly")
        self.add_physical(pmlz, "pmlz")

        self.add_physical(pmlxy, "pmlxy")
        self.add_physical(pmlyz, "pmlyz")
        self.add_physical(pmlxz, "pmlxz")

        self.add_physical(pml3, "pmlxyz")


class Scatt3D(Simulation):
    def __init__(
        self,
        geometry,
        epsilon,
        mu,
        source,
        boundary_conditions={},
        degree=1,
        mat_degree=1,
        pml_stretch=1 - 1j,
    ):
        assert isinstance(geometry, BoxPML3D)
        assert source.dim == 3
        function_space = ComplexFunctionSpace(geometry.mesh, "N1curl", degree)
        pmls = []
        pml_names = []
        for direction in ["x", "y", "z", "xy", "yz", "xz", "xyz"]:
            pml_name = f"pml{direction}"
            pml_names.append(pml_name)
            pmls.append(
                PML(
                    direction,
                    stretch=pml_stretch,
                    matched_domain="box",
                    applied_domain=pml_name,
                )
            )

        epsilon_coeff = Coefficient(
            epsilon,
            geometry,
            pmls=pmls,
            degree=mat_degree,
            dim=3,
        )
        mu_coeff = Coefficient(
            mu,
            geometry,
            pmls=pmls,
            degree=mat_degree,
            dim=3,
        )

        coefficients = epsilon_coeff, mu_coeff
        no_source_domains = ["box"] + pml_names
        source_domains = [
            dom for dom in geometry.domains if dom not in no_source_domains
        ]

        formulation = Maxwell3D(
            geometry,
            coefficients,
            function_space,
            source=source,
            source_domains=source_domains,
            reference="box",
            boundary_conditions=boundary_conditions,
        )

        super().__init__(geometry, formulation)

    def solve_system(self, again=False):
        E = super().solve_system(again=again, vector_function=False)
        self.solution = {}
        self.solution["diffracted"] = E
        self.solution["total"] = E + self.source.expression
        return E



import codecs
import os

from setuptools import find_packages, setup


def read(fname):
    return codecs.open(os.path.join(base_dir, fname), encoding="utf-8").read()


# https://packaging.python.org/single_source_version/
base_dir = os.path.abspath(os.path.dirname(__file__))
about = {}
with open(os.path.join(base_dir, "gyptis", "__about__.py"), "rb") as f:
    exec(f.read(), about)


required = [
    "scipy",
    "dolfin-adjoint",
    "meshio",
    "gmsh",
    "h5py",
    "nlopt",
    "numpy",
    "simpy",
]

_classifiers = [
    about["__status__"],
    about["__license__"],
    about["__operating_system__"],
]
_classifiers += about["__programming_language__"]
_classifiers += about["__topic__"]

setup(
    name="gyptis",
    version=about["__version__"],
    author=about["__author__"],
    author_email=about["__author_email__"],
    packages=find_packages(),
    description=about["__description__"],
    long_description=read("README.rst"),
    long_description_content_type="text/x-rst",
    url=about["__website__"],
    project_urls={"Documentation": about["__website__"]},
    license=about["__license__"],
    platforms="any",
    include_package_data=True,
    install_requires=required,
    extras_require={},
    classifiers=_classifiers,
)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Author: Benjamin Vial
# License: MIT

import sys

import numpy as np

from . import dolfin


def array2function(a, A):
    """Convert a numpy array to a fenics Function.

    Parameters
    ----------
    a : numpy array
        The array to convert.
    A : FunctionSpace
        The function space to interpolate on.

    Returns
    -------
    Function
        The converted array.

    """
    u = dolfin.Function(A)
    u.vector().set_local(a)
    dolfin.as_backend_type(u.vector()).update_ghost_values()
    u.vector().apply("insert")
    return u


def function2array(f):
    """Convert a fenics Function to a numpy array.

    Parameters
    ----------
    f : Function
        The function to convert.

    Returns
    -------
    numpy array
        The converted function.

    """
    return f.vector().get_local()


def list_time():
    return dolfin.list_timings(dolfin.TimingClear.clear, [dolfin.TimingType.wall])


def get_coordinates(A):
    n = A.dim()
    d = A.mesh().geometry().dim()
    dof_coordinates = A.tabulate_dof_coordinates()
    dof_coordinates.resize((n, d))
    return dof_coordinates


def mpi_print(s):
    if dolfin.MPI.rank(dolfin.MPI.comm_world) == 0:
        print(s)
        sys.stdout.flush()


def matfmt(m, ndigit=4, extra_space=0, cplx=False):
    def printim(y):
        return f"{sgn(float(y))}{abs(float(y))}j"

    def sgn(u):
        if np.sign(u) == -1:
            return "-"
        else:
            return "+"

    dim = len(m[0])

    pad = " " * extra_space

    if cplx:
        m = [[_.real, _.imag] for _ in np.ravel(m)]

    a = [f"%.{ndigit}f" % elem for elem in np.ravel(m)]

    if dim == 3:
        if cplx:
            b = f"""[{a[0]}{printim(a[1])} {a[2]}{printim(a[3])} 
            {a[4]}{printim(a[5])}] \n{pad}[{a[6]}{printim(a[7])} 
            {a[8]}{printim(a[9])} {a[10]}{printim(a[11])}] 
            \n{pad}[{a[12]}{printim(a[13])} {a[14]}{printim(a[15])} 
            {a[16]}{printim(a[17])}]
            """

        else:
            b = f"""[{a[0]} {a[1]} {a[2]}] \n{pad}[{a[3]} {a[4]} {a[5]}]
             \n{pad}[{a[6]} {a[7]} {a[8]}]
             """
    else:
        if cplx:
            b = f"""[{a[0]}{printim(a[1])} {a[2]}{printim(a[3])}] 
            \n{pad}[{a[4]}{printim(a[5])} {a[6]}{printim(a[7])}]
            """
        else:
            b = f"[{a[0]} {a[1]}] \n{pad}[{a[2]} {a[3]}]"
    return b


def matprint(*args, **kwargs):
    mpi_print(matfmt(*args, **kwargs))


def rot_matrix_2d(t):
    return np.array([[np.sin(t), -np.cos(t), 0], [np.cos(t), np.sin(t), 0], [0, 0, 1]])


def tanh(x):
    return (dolfin.exp(2 * x) - 1) / (dolfin.exp(2 * x) + 1)


def _translation_matrix(t):
    M = [1.0, 0.0, 0.0, 0.0, 0.0, 1.0, 0.0, 0.0, 0.0, 0.0, 1.0, 0.0, 0.0, 0.0, 0.0, 1.0]
    M[3], M[7], M[11] = t
    return M



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Author: Benjamin Vial
# License: MIT


import numpy as np
from scipy.optimize import OptimizeResult, minimize

from . import dolfin as df
from .complex import *
from .helpers import *
from .materials import tensor_const


def simp(a, s_min=1, s_max=2, p=1):
    """Solid isotropic material with penalisation (SIMP)"""
    return s_min + (s_max - s_min) * a ** p


def tanh(x):
    return (df.exp(2 * x) - 1) / (df.exp(2 * x) + 1)


def projection(a, beta=1, nu=0.5):
    """Projection operator."""
    return (tanh(beta * nu) + tanh(beta * (a - nu))) / (
        tanh(beta * nu) + tanh(beta * (1 - nu))
    )


def filtering(a, rfilt=0, solver="iterative", function_space=None, order=1, dim=2):
    assert solver in ["direct", "iterative"]
    if np.all(rfilt == 0):
        return a
    else:
        mesh = a.function_space().mesh()
        dim = mesh.ufl_domain().geometric_dimension()
        F = function_space or df.FunctionSpace(mesh, "CG", order)
        bcs = []
        af = df.TrialFunction(F)
        vf = df.TestFunction(F)
        if hasattr(rfilt, "shape"):
            if np.shape(rfilt) in [(2, 2), (3, 3)]:
                rfilt_ = tensor_const(rfilt, dim=dim, real=True)
            else:
                raise ValueError("Wrong shape for rfilt")
        else:
            rfilt_ = df.Constant(rfilt)
        a_ = (
            df.inner(rfilt_ * df.grad(af), rfilt_ * df.grad(vf)) * df.dx
            + df.inner(af, vf) * df.dx
        )
        L_ = df.inner(a, vf) * df.dx
        af = df.Function(F, name="Filtered density")
        if solver == "direct":
            df.solve(a_ == L_, af, bcs)
            # solve(
            #     Ff == 0, af, bcs, solver_parameters=solver_parameters,
            # )
        else:
            solver = df.KrylovSolver("cg", "jacobi")
            # solver.parameters['relative_tolerance'] = 1e-3
            A = df.assemble(a_)
            b = df.assemble(L_)
            solver.solve(A, af.vector(), b)

        return af


def derivative(f, x, ctrl_space=None, array=True):
    dfdx = df.compute_gradient(f, df.Control(x))
    if ctrl_space is not None:
        dfdx = project(dfdx, ctrl_space)
    if array:
        return function2array(dfdx)
    else:
        return dfdx


class OptimFunction:
    def __init__(self, fun, stop_val=None, normalization=1):
        self.fun_in = fun
        self.fun_value = None
        self.stop_val = stop_val
        self.normalization = normalization

    def fun(self, x, *args, **kwargs):
        if self.fun_value is not None:
            if self.stop_val is not None:
                if self.fun_value < self.stop_val:
                    raise ValueError(
                        f"Minimum value {self.stop_val} reached, current objective function = {self.fun_value}"
                    )
        obj, grad = self.fun_in(x, *args, **kwargs)
        obj /= self.normalization
        grad /= self.normalization
        self.fun_value = obj
        self.x = x
        return obj, grad


def scipy_minimize(
    f,
    x,
    bounds,
    maxiter=50,
    maxfun=100,
    tol=1e-6,
    args=(),
    stop_val=None,
    normalization=1,
):

    optfun = OptimFunction(fun=f, stop_val=stop_val, normalization=normalization)
    opt = OptimizeResult()
    opt.x = x
    opt.fun = None
    try:
        opt = minimize(
            optfun.fun,
            x,
            tol=tol,
            args=args,
            method="L-BFGS-B",
            jac=True,
            bounds=bounds,
            options={"maxiter": maxiter, "maxfun": maxfun},
        )
    except Exception as e:
        print(e)
        opt.x = optfun.x
        opt.fun = optfun.fun_value

    return opt


def transfer_sub_mesh(x, geometry, source_space, target_space, subdomain):
    markers = geometry.markers["triangle"]
    domains = geometry.subdomains["surfaces"]
    a = df.Function(source_space)
    mdes = markers.where_equal(domains[subdomain])
    a = function2array(a)
    a[mdes] = function2array(project(x, target_space))
    ctrl = array2function(a, source_space)
    return ctrl



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Author: Benjamin Vial
# License: MIT


__all__ = ["BoxPML", "Layered", "Scattering", "Grating"]

from .grating2d import Grating2D, Layered2D
from .grating3d import Grating3D, Layered3D
from .scattering2d import BoxPML2D, Scatt2D
from .scattering3d import BoxPML3D, Scatt3D


class BoxPML:
    def __new__(self, dim=3, *args, **kwargs):
        if dim not in [2, 3]:
            raise ValueError("dimension must be 2 or 3")
        if dim == 3:
            return BoxPML3D(*args, **kwargs)
        else:
            return BoxPML2D(*args, **kwargs)


class Layered:
    def __new__(self, dim=3, *args, **kwargs):
        if dim not in [2, 3]:
            raise ValueError("dimension must be 2 or 3")
        if dim == 3:
            return Layered3D(*args, **kwargs)
        else:
            return Layered2D(*args, **kwargs)


class Scattering:
    def __new__(self, *args, **kwargs):
        geom = kwargs.get("geometry") or args[0]
        if geom.dim not in [2, 3]:
            raise ValueError("dimension must be 2 or 3")
        if geom.dim == 3:
            return Scatt3D(*args, **kwargs)
        else:
            return Scatt2D(*args, **kwargs)


class Grating:
    def __new__(self, *args, **kwargs):
        geom = kwargs.get("geometry") or args[0]
        if geom.dim not in [2, 3]:
            raise ValueError("dimension must be 2 or 3")
        if geom.dim == 3:
            return Grating3D(*args, **kwargs)
        else:
            return Grating2D(*args, **kwargs)



import os

import dolfin

from .__about__ import (
    __author__,
    __author_email__,
    __copyright__,
    __description__,
    __license__,
    __status__,
    __version__,
    __website__,
)

__doc__ = __description__


if os.environ.get("GYPTIS_ADJOINT") is not None:
    import dolfin_adjoint

    ADJOINT = True
    dolfin.__dict__.update(dolfin_adjoint.__dict__)
    dolfin.__spec__.name = "dolfin"
    del dolfin_adjoint

    ## prevents AttributeError: module 'fenics_adjoint.types.function'
    ## has no attribute 'function' when writing solution to files with <<
    dolfin.function.function = dolfin.function
else:
    ADJOINT = False

dolfin.parameters["reorder_dofs_serial"] = False


dolfin.PETScOptions.set("petsc_prealloc", "200")
dolfin.PETScOptions.set("ksp_type", "preonly")
dolfin.PETScOptions.set("pc_type", "lu")
# dolfin.PETScOptions.set('pc_factor_mat_solver_type', 'mumps')
# dolfin.PETScOptions.set('sub_pc_type', 'lu')
# dolfin.PETScOptions.set('pc_asm_overlap', '10')
# dolfin.PETScOptions.set('ksp_type', 'gmres')

# dolfin.PETScOptions.set('petsc_prealloc', '100')
# dolfin.PETScOptions.set('ksp_rtol', '1.e-12')
# dolfin.PETScOptions.set('ksp_type', 'gmres')
# dolfin.PETScOptions.set('ksp_gmres_restart', '100')
# dolfin.PETScOptions.set('pc_type', 'jacobi')
# dolfin.PETScOptions.set('pc_factor_levels', '3')
# dolfin.PETScOptions.set('mat_mumps_icntl_14', 40)
# dolfin.PETScOptions.set('mat_mumps_icntl_35', 2)
# dolfin.PETScOptions.set('mat_mumps_icntl_36', 1)
# dolfin.PETScOptions.set('mat_mumps_icntl_24', 1)

# dolfin.set_log_level(1)

from .api import *
